local flowerCount = 12

local shipImg = actor.LoadSprite("image/ship.png", { 64, 64 })

local LEFT   = 1
local RIGHT  = 2
local UP     = 3
local DOWN   = 4

actor.Cage("Blocker", 0.6, 0.5)
actor.SimpleCollide("Blocker", "Player", actor.Blocker)

local animFPS = 64

local function AnimatePlayer(player, stop)
	local start = eapi.GetFrame(player.tile)
	local time = math.abs(stop - start) / animFPS
	eapi.AnimateFrame(player.tile, eapi.ANIM_CLAMP, start, stop, nil, time)
end

local function Animate(player)
	local vel = eapi.GetVel(player.body)
	if vel.x == 0 then
		AnimatePlayer(player, 16)
	elseif vel.x > 0 then
		AnimatePlayer(player, 8)
	elseif vel.x < 0 then
		AnimatePlayer(player, 24)
	end
end

local function GetDirection(moves)
	return { x = moves[RIGHT] - moves[LEFT], y = moves[UP] - moves[DOWN] }
end

local function SetSpeed(player, speed)
	if speed then player.speed = speed end
	eapi.SetVel(player.body, vector.Normalize(player.vel, player.speed))
end

local function EnableInput(player, Fn)
	input.Bind("Left", false, Fn(player, LEFT))
	input.Bind("Right", false, Fn(player, RIGHT))
	input.Bind("Up", false, Fn(player, UP))
	input.Bind("Down", false, Fn(player, DOWN))
	return player
end

local function Move(player, axis)
	return function(keyDown)
		player.moves[axis] = (keyDown and 1) or 0
		player.vel = GetDirection(player.moves)
		SetSpeed(player)
		Animate(player)
	end
end

local function RunFlowers(obj, disable)
	local white = util.Gray(1.0)
	local dT = player.patternTime / flowerCount
	if obj.flowerTimer then
		eapi.CancelTimer(obj.flowerTimer)
		obj.flowerTimer = nil
	end
	local function Run()
		local tile = obj.flowers[obj.flowerIndex]
		if tile then
			eapi.AnimateColor(tile, eapi.ANIM_CLAMP, white, dT, 0)
			obj.flowerTimer = eapi.AddTimer(obj.body, dT, Run)
			obj.flowerIndex = obj.flowerIndex + 1
		else
			obj.flowerTimer = nil
		end
	end
	if player.shrinkEnable then Run() end
end

local function FirstMove(player, axis)
	return function(keyDown)
		Move(player, axis)(keyDown)
		util.Map(eapi.Destroy, player.text)
		EnableInput(player, Move)
		RunFlowers(player)
		bomb.ResetTime()
		timing.Start()
	end
end

local heartImg = actor.LoadSprite("image/heart.png", { 256, 256 })

local function AddHeart(obj)
	local speed = 0.2
	local animType = eapi.ANIM_LOOP
	local size = { x = 16, y = 16 }
	local offset = { x = -8, y = -3 }
	obj.heart = eapi.NewTile(obj.body, offset, size, heartImg, obj.z + 0.1)
	eapi.SetColor(obj.heart, { r = 0.25, g = 0.05, b = 0, a = 0.25 })
	local dstSize = { x = 8, y = 8 }
	local dstOffset = { x = -4, y = 1 }
	eapi.AnimateSize(obj.heart, animType, dstSize, speed, 0)
	eapi.AnimatePos(obj.heart, animType, dstOffset, speed, 0)
	local dstColor = { r = 0.75, g = 0.15, b = 0, a = 1.0 }
	eapi.AnimateColor(obj.heart, animType, dstColor, speed, 0)
end

local function Trail(obj)
	local ttl = 1.0
	local z = obj.z - 0.05
	local srcColor = { r = 1.0, g = 1.0, b = 1.0, a = 0.1 }
	local dstColor = { r = 1.0, g = 0.0, b = 0.0, a = 0.0 }
	local dstPos = { x = -128, y = -128 }
	local dstSize = { x = 256, y = 256 }
	local offset = { x = -32, y = -32 }
	local vel = { x = 0, y = -50 }
	local function Emit()
		local pos = eapi.GetPos(obj.body)
		local body = eapi.NewBody(gameWorld, pos)
		local tile = eapi.NewTile(body, offset, nil, shipImg, z)

		eapi.SetVel(body, vel)
		eapi.SetColor(tile, srcColor)
		eapi.AnimateSize(tile, eapi.ANIM_CLAMP, dstSize, ttl, 0)
		eapi.AnimatePos(tile, eapi.ANIM_CLAMP, dstPos, ttl, 0)
		eapi.AnimateColor(tile, eapi.ANIM_CLAMP, dstColor, ttl, 0)
		eapi.SetFrame(tile, eapi.GetFrame(obj.tile))
		eapi.AddTimer(obj.body, 0.01, Emit)
		util.DelayedDestroy(body, ttl)
	end
	Emit()
end

local function CreateFlowers(obj)
	local index = 1
	local z = obj.z - 1.0
	local arm = { x = 0, y = 32 }
	local size = { x = 16, y = 16 }
	local offset = { x = -8, y = -8 }
	local sprite = explode.petalImg
	for angle = 0, 359, 360 / flowerCount do
		local pos = vector.Rotate(arm, angle)
		local body = eapi.NewBody(obj.body, pos)
		local tile = eapi.NewTile(body, offset, size, sprite, z)
		eapi.SetColor(tile, util.invisible)
		util.AnimateRotation(tile, 1)
		eapi.SetFrame(tile, 29)
		obj.flowers[index] = tile
		index = index + 1
	end
end

local function Create()
	local obj = { z = 0,
		      speed = 300,
		      flowers = { },
		      flowerIndex = 1,
		      class = "Player",
		      sprite = shipImg,
		      vel = vector.null,		      
		      moves = { 0, 0, 0, 0 },
		      pos = { x = 0, y = -200 },
		      offset = { x = -32, y = -32 },
		      bb = { l = -4, r = 4, b = 2, t = 10 }, }	
	local text = "\"Arrow Keys\" - Evade"
	ribbons.ShowHelpText(obj, text)
	obj = actor.Create(obj)
	eapi.SetFrame(obj.tile, 16)
	eapi.PlayMusic("sound/music.ogg", nil, 1.0, 1.0)
	EnableInput(obj, FirstMove)
	CreateFlowers(obj)
	player.obj = obj
	AddHeart(obj)
	Trail(obj)
end

local function QuickFade(tile)
	eapi.SetColor(tile, util.invisible)
end

local function ShrinkCircle()
	local obj = player.obj
	util.Map(QuickFade, obj.flowers)	
	obj.flowerIndex = 1
	RunFlowers(obj)
end

local function EnableShrink(state)
	player.shrinkEnable = state
	ShrinkCircle()
end

player = {
	Create = Create,
	patternTime = 6.28,
	ShrinkCircle = ShrinkCircle,
	EnableShrink = EnableShrink,
	shrinkEnable = true,
}
return player
